import matplotlib.pyplot as plt
from copy import deepcopy
import numpy as np
import pandas as pd
import warnings

from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis, get_expert_decision
from lhcsmapi.dbsignal.SignalProcessing import SignalProcessing


def find_time_of_i_a_i_ref_deviation(i_ref_df, i_a_df):
    i_ref_abs_df = abs(i_ref_df)
    i_a_abs_df = abs(i_a_df)
    # re-interpolate both signals
    idx = i_ref_abs_df.index.union(i_a_abs_df.index)
    signal_comparison_df = i_ref_abs_df.reindex(idx).interpolate('index')
    signal_comparison_df = signal_comparison_df.join(i_a_abs_df).dropna()
    signal_comparison_df[(signal_comparison_df.index > -1) & (signal_comparison_df.index < i_ref_df.idxmax().values[0])]

    # find maximum difference between the time of maximum I_REF value and 1 second before
    t_i_ref_max = i_ref_abs_df.idxmax().values[0]
    diff_df = abs(signal_comparison_df['STATUS.I_REF'] - signal_comparison_df['IAB.I_A'])
    max_diff = diff_df[(diff_df.index > (t_i_ref_max - 1)) & (diff_df.index < t_i_ref_max)].max()

    # find time of start of deviation from maximum difference between maximum value of I_REF and 0 (FGC timestamp)
    mask = (diff_df.index > t_i_ref_max) & (diff_df.index < 0) & (diff_df.values > max_diff)

    return diff_df[mask].idxmin()


def flip_signal(u_qs0_ref):
    # # if there is only a negative peak, then flip the signal
    # # if there are two peaks and the first is negative, then flip the signal
    # # otherwise, do nothing
    # # added the case when U_QS0 is saturated only to negative!
    if (all(u_qs0_ref.max() < 0.2) and all(u_qs0_ref.min() < -0.2)):
        return -deepcopy(u_qs0_ref)
    elif (u_qs0_ref.max().values[0] > 0.2) & (u_qs0_ref.min().values[0] < -0.2) & (
            u_qs0_ref.idxmin().values[0] < u_qs0_ref.idxmax().values[0]):
        return -deepcopy(u_qs0_ref)
    else:
        return deepcopy(u_qs0_ref)


def plot_pc_current(i_a_df, i_meas_df, i_ref_df, title, figsize=(20, 7), xlim=None, is_zoom=False, t_deviation=None):
    fig, ax = plt.subplots(figsize=figsize)
    if not i_meas_df.empty:
        i_meas_df.plot(ax=ax, title=title)
    if not i_ref_df.empty:
        i_ref_df.plot(ax=ax)
    if not i_a_df.empty:
        i_a_df.plot(ax=ax)
    ax.grid(True)
    ax.title.set_size(20)
    ax.tick_params(labelsize=15)
    ax.set_xlabel("time, [s]", fontsize=15)
    ax.set_ylabel("I, [A]", fontsize=15)

    if xlim is not None:
        ax.set_xlim(xlim)

    if is_zoom and not i_meas_df.empty:
        max_ylim = i_meas_df.max().values[0]
        min_ylim = i_meas_df.min().values[0]
        if abs(max_ylim) > abs(min_ylim):
            ax.set_ylim([max_ylim - 10, max_ylim + 10])
        else:
            ax.set_ylim([min_ylim + 10, min_ylim - 10])

    if t_deviation is not None:
        ax.axvline(t_deviation, ls='--')

    plt.show()


class _600ACircuitAnalysis(CircuitAnalysis):

    # PC
    def analyze_i_meas_pc(self, circuit_name, timestamp_fgc, i_meas_df, i_ref_df, i_a_df):
        # Ramp Rate and Plateau Duration
        if i_meas_df.empty:
            self.results_table['Ramp Rate'] = np.nan
            self.results_table['Plateau Duration'] = np.nan
            warnings.warn('I_MEAS signal is empty, analysis skipped')
        else:
            idx = max([i_meas_df.idxmin().values[0], i_meas_df.idxmax().values[0]])
            i_meas_df_der = i_meas_df[i_meas_df.index < idx]
            SignalProcessing.calculate_time_derivative_of_signals(i_meas_df_der, i_meas_df_der.index, ['STATUS.I_MEAS'],
                                                                  ['dSTATUS.I_MEAS'])

            didt_max = i_meas_df_der['dSTATUS.I_MEAS'].max()
            if abs(didt_max) > 0.1:
                self.results_table['Ramp Rate'] = didt_max
                self.results_table['Plateau Duration'] = 0
            else:
                self.results_table['Ramp Rate'] = 0
                warnings.warn('Calculation of Plateau Duration is not yet implemented.')

        # I_MEAS MIITs
        if (not i_ref_df.empty) and (not i_a_df.empty):
            t_deviation = find_time_of_i_a_i_ref_deviation(i_ref_df, i_a_df)
            i_meas_pos_df = i_meas_df[i_meas_df.index > t_deviation]
            miits_i_meas = np.trapz(i_meas_pos_df.values.reshape(len(i_meas_pos_df.index)) ** 2,
                                    i_meas_pos_df.index) * 1e-6
            self.results_table['I_MEAS MIITs'] = round(miits_i_meas, 3)
            print('The I_MEAS(t) MIITs is %s MA^2s.' % round(miits_i_meas, 3))
            self.results_table['Quench Current'] = round(i_meas_pos_df['STATUS.I_MEAS'].iloc[0], 1)
            print('Quench Current is %s A.' % round(i_meas_pos_df['STATUS.I_MEAS'].iloc[0], 1))
        else:
            self.results_table['I_MEAS MIITs'] = np.nan
            warnings.warn('Some signals are empty, calculation of I_MEAS MIITs skipped')
            return None

        title = "%s, %s: I_MEAS(t), I_REF(t), I_A(t)" % (Time.to_string_short(timestamp_fgc), circuit_name)
        plot_pc_current(i_a_df, i_meas_df, i_ref_df, title)
        plot_pc_current(i_a_df, i_meas_df, i_ref_df, title, figsize=(15, 15), xlim=[t_deviation - 0.01, 0.01],
                        is_zoom=True, t_deviation=t_deviation)

    def analyze_i_earth_pc(self, circuit_name, timestamp_fgc, i_earth_df):
        title = "%s, %s: I_EARTH(t)" % (Time.to_string_short(timestamp_fgc), circuit_name)

        fig, ax = plt.subplots(figsize=(15, 7.5))
        if not i_earth_df.empty:
            i_earth_df.plot(ax=ax, grid=True, title=title)
            ax.legend(loc='lower left')
            ax.title.set_size(20)
            ax.set_xlabel("time, [s]", fontsize=15)
            ax.set_ylabel("IEARTH, [A]", fontsize=15, color='C0')
            ax.tick_params(labelsize=15)
            plt.show()

            i_earth_max = i_earth_df.max().values[0] if i_earth_df.max().values[0] > abs(
                i_earth_df.min().values[0]) else i_earth_df.min().values[0]
            self.results_table['Earth Current'] = round(i_earth_max, 3)
            print('The maximum earth current is %s A.' % round(i_earth_max, 3))
        else:
            self.results_table['Earth Current'] = np.nan
            warnings.warn('Earth current signal empty!')

    # EE
    def analyze_u_dump_res_ee(self, circuit_name, timestamp_fgc, i_meas_df, u_dump_res_df):

        fig, ax = plt.subplots(figsize=(20, 7))

        title = "%s, %s: I_EARTH(t)" % (Time.to_string_short(timestamp_fgc), circuit_name)
        if not i_meas_df.empty:
            i_meas_df.plot(ax=ax, grid=True, color='C0', title=title)
            ax.legend(loc='lower left')
            ax.title.set_size(20)
            ax.set_xlabel("time, [s]", fontsize=15)
            ax.set_ylabel("I_MEAS, [A]", fontsize=15, color='C0')
            ax.tick_params(labelsize=15)

        if not u_dump_res_df.empty:
            ax1 = ax.twinx()
            u_dump_res_df.plot(ax=ax1, color='C2')
            ax1.legend(loc='upper right')
            ax1.set_ylabel("U_DUMP_RES, [V]", fontsize=15, color='C2')
            ax1.tick_params(labelsize=15)
            ax1.set_xlim(-1, 2)
            plt.show()
            u_dump_res_max = u_dump_res_df.max().values[0] if u_dump_res_df.max().values[0] > abs(
                u_dump_res_df.min().values[0]) else u_dump_res_df.min().values[0]
            self.results_table['EE max voltage'] = round(u_dump_res_max, 3)
            print('The maximum EE voltage is %s V.' % round(u_dump_res_max, 3))
        else:
            self.results_table['EE max voltage'] = np.nan
            warnings.warn('EE voltage signal empty!')

    # QPS
    def analyze_qds(self, circuit_name, timestamp_qps, i_dcct_df, i_didt_df, u_diff_df, u_res_df):
        title = "%s, %s: U_RES(t), U_DIFF(t), I_DCCT(t), I_DIDT(t)" % (
            Time.to_string_short(timestamp_qps), circuit_name)

        fig, ax = plt.subplots()
        fig.suptitle(title, fontsize=20)
        ax.set_xlabel("time, [s]", fontsize=15)
        ax.tick_params(labelsize=15)

        if not u_diff_df.empty:
            u_diff_df.plot(figsize=(20, 7), ax=ax, color='C0')
            ax.legend(loc='upper left')
            ax.set_ylim([-10, 10])
            ax.set_ylabel("U_DIFF, [V]", fontsize=15, color='C0')

        if not u_res_df.empty:
            ax3 = ax.twinx()
            ax3.spines["right"].set_position(("axes", -0.11))
            u_res_df.plot(ax=ax3, color='C1')
            ax3.set_ylabel("U_RES, [V]", fontsize=15, color='C1')
            ax3.legend(loc='lower left')
            ax3.set_ylim([-0.25, 0.25])
            ax3.tick_params(labelsize=15)
            ax3.tick_params(axis="y", direction="in", pad=-50)
            ax3.yaxis.set_label_coords(-0.175, 0.5)

        if not i_dcct_df.empty:
            ax1 = ax.twinx()
            i_dcct_df.plot(ax=ax1, color='C2')
            ax1.legend(loc='upper right')
            ax1.tick_params(labelsize=15)
            ax1.set_ylabel("I_DCCT, [A]", fontsize=15, color='C2')

        if not i_didt_df.empty:
            ax2 = ax.twinx()
            ax2.spines["right"].set_position(("axes", 1.1))
            i_didt_df.plot(ax=ax2, grid=True, color='C3')
            ax2.legend(loc='lower right')
            ax2.tick_params(labelsize=15)
            ax2.set_ylabel("I_DIDT, [A/s]", fontsize=15, color='C3')

        ax.set_xlim([-0.1, 0.1])

        plt.show()

        # Get expert input on reason
        self.results_table['Reason'] = get_expert_decision('Reason for FPA: ', ['quench', 'coupling', 'other'])

        # Get expert input on quench origin
        self.results_table['Quench Origin'] = get_expert_decision('Origin of a quench', ['magnet', 'busbar', 'other'])

    def calculate_u_res_slope(self, circuit_name, timestamp_qps, u_res_df):
        if u_res_df.empty:
            raise ValueError('U_RES signal is empty')

        # Skip first ten points to avoid initial noise
        u_res_slice_df = u_res_df[u_res_df.index > u_res_df.index[10]]
        # Take absolute value
        u_res_flip_df = flip_signal(u_res_slice_df)

        # Find first index of argmax (0.2 V)
        t_u_max = u_res_flip_df[u_res_flip_df >= 0.2].dropna().index[0]
        v_u_max = u_res_df[u_res_df.index == t_u_max].dropna().values[0]

        # Find index of voltage equal to 0.025 V
        temp = u_res_flip_df[(u_res_flip_df.index < t_u_max)]
        t_u_min = temp[(temp.values < 0.025)].index[-1]
        v_u_min = u_res_df[u_res_df.index == t_u_min].dropna().values[0][0]

        # Calculate dU_RES/dt
        d_u_res_dt = (v_u_max - v_u_min) / (t_u_max - t_u_min)

        # self.results_table['dUres/dt'] = d_u_res_dt[0]

        title = "%s, %s: U_RES(t), U_DIFF(t), I_DCCT(t), I_DIDT(t)" % (
            Time.to_string_short(timestamp_qps), circuit_name)

        ax = u_res_df[u_res_df.index > u_res_df.index[10]].plot(figsize=(15, 7.5), grid=True, title=title)
        ax.legend(loc='lower left')
        ax.title.set_size(20)
        ax.set_xlabel("time, [s]", fontsize=15)
        ax.set_ylabel("U_RES, [V]", fontsize=15, color='C0')
        ax.tick_params(labelsize=15)
        slope_df = pd.DataFrame(data=[v_u_min, v_u_max[0]], index=[t_u_min, t_u_max], columns=['dU_RES/dt'])
        slope_df.plot(ax=ax, grid=True)
        ax.legend([u_res_df.columns[0], slope_df.columns[0]])
        plt.show()

        print('The slope of the U_RES signal is %s V/s.' % round(d_u_res_dt[0], 3))

    def create_mp3_results_table(self):
        columns = ['Circuit Name', 'Family', 'Period', 'Date', 'Time', 'flagged in APEX', 'Quench Current',
                   'DCCT MIITs', 'Reason', 'Pivot', 'Quench Origin',
                   'Ramp Rate', 'Plateau Duration', 'dUres/dt', 'EE max voltage', 'Earth Current', 'Coupling',
                   'Comment', 'Person who analyzed', 'Reference Doc']
        mp3_table = pd.DataFrame(columns=columns, index=[0])

        mp3_table['Circuit Name'] = self.results_table['Circuit Name']
        mp3_table['Family'] = self.results_table['Family']
        mp3_table['Period'] = self.results_table['Period']
        mp3_table['Date'] = self.results_table['Time']
        mp3_table['flagged in APEX'] = self.results_table['flagged in APEX']
        mp3_table['Quench Current'] = self.results_table['Quench Current']
        mp3_table['DCCT MIITs'] = self.results_table['I_MEAS MIITs']
        mp3_table['Reason'] = self.results_table['Reason']
        mp3_table['Pivot'] = self.results_table['Pivot']
        mp3_table['Quench Origin'] = self.results_table['Quench Origin']
        mp3_table['Ramp Rate'] = self.results_table['Ramp Rate']
        mp3_table['Plateau Duration'] = self.results_table['Plateau Duration']
        mp3_table['dUres/dt'] = self.results_table['dUres/dt']
        mp3_table['EE max voltage'] = self.results_table['EE max voltage']
        mp3_table['Coupling'] = self.results_table['Coupling']
        mp3_table['Comment'] = self.results_table['Comment']
        mp3_table['Person who analyzed'] = self.results_table['performed_by']
        mp3_table['Reference Doc'] = self.results_table['Reference Doc']

        return mp3_table

